{-----------------------------------------------------------------------------

 CF_Play : main playfield definitions of the game ChiFu.

 Consists of:

 Types:
 -TCoord    : quick coordinate reference
 -TSortCode : all sorts of stones possible
 -TStone    : single loose stone reference
 -TBlock    : block definition
   Variables:
   -Position : position of block in field
   -Stuck    : whether block is stuck in field

   Routines:
   -MoveDown     : moves the block one down
   -MoveLeft     : moves the block one left
   -MoveRight    : moves the block one Right
   -TurnUp       : turns the block upwards
   -TurnDown     : turns the block downwards
   -MakeNewBlock : makes a new block
   -GetStone     : gets the stone at the specified StoneNr

 -TField : playfield definition
    Variables:

    Routines:
    -InsertBlock  : inserts the Block in the field
    -Occupied     : tells wether the given coordinate is occupied
    -ClearField   : clears the field
    -GetStone     : gets the stone at the specified coordinates
    -CheckRows    : returns how many stones were deleted in full rows
    -MakeFullRows : makes full rows in the bottom of the field

 Constants:
 -JokerStoneNr : standard stone nr. of the joker

 Variables:
 -Field    : the playfield
 -Block    : the moving block

 Routines:

 Defaults:
 -Initializes the random generator
 -Clears the playfield

-----------------------------------------------------------------------------}
unit CF_Play;

interface

uses
  CF_Set;

const
  JokerStoneNr : byte = 0;

type
  { Quick qoordinate reference }
  TCoord = record
             X,
             Y : integer;
           end; { record }

  { All sorts of stones possible }
  TSortCode = (NoStone, NormalStone, DeletedStone);

  { Single loose stone reference }
  TStone = record
             Coord   : TCoord;
             Variety : byte;
             Sortof  : TSortCode;
           end; { record }

  { Block definition }
  TBlock = object
             public
               Position : TCoord;   { Position of block in field }
               Stuck    : boolean;  { Whether block is stuck in field }

               { Moves the block one down }
               procedure MoveDown;

               { Moves the block one left }
               procedure MoveLeft;

               { Moves the block one Right }
               procedure MoveRight;

               { Turns the block upwards }
               procedure TurnUp;

               { Turns the block downwards }
               procedure TurnDown;

               { Makes a new block }
               procedure MakeNewBlock;

               { Gets the stone at the specified StoneNr }
               procedure GetStone (StoneNr : byte; var Stone : TStone);

             private
               Stones : array [1..6] of TStone;  { All stones in block }

               { Makes new varieties in block }
               procedure MakeNewVarieties;

               { Gives the block a new coordinate }
               procedure MakeNewCoord;


           end; { object }

  { Playfield definition }
  TField = object
             public

               { Inserts the Block in the field }
               procedure InsertBlock;

               { Tells wether the given coordinate is occupied }
               function Occupied (Coord : TCoord) : boolean;

               { Clears the field }
               procedure ClearField;

               { Gets the stone at the specified coordinates }
               procedure GetStone (Coord : TCoord; var Stone : TStone);

               { Returns how many stones were deleted in full rows }
               function CheckRows : word;

               { Makes full rows in the bottom of the field }
               procedure MakeFullRows;

             private
               AllStones : array [1..11, 1..18] of TStone;

           end; { object }

var
  Field : TField;   { The playfield }
  Block : TBlock;   { Loose block to move in field }

implementation

{ Block member functions------------------------------}

{ Turns the block one upwards }
procedure TBlock.TurnUp;

var
  CarryVariety,
  StoneNr    : byte;

begin
  { Remember variety to carry over to other end }
  CarryVariety := Stones [1].Variety;

  { And copy all varieties one down }
  if Settings [BlockLength] <> 1 then begin
    for StoneNr := 1 to Settings [BlockLength] - 1 do begin
      Stones [StoneNr].Variety := Stones [StoneNr + 1].Variety;
    end; { for }
  end; { then }

  { And put in the carried Variety }
  Stones [Settings [BlockLength]].Variety := CarryVariety;
end; { TurnUp }


{ Turns the block one downwards }
procedure TBlock.TurnDown;

var
  CarryVariety,
  StoneNr    : byte;

begin
  { Remember variety to carry over to other end }
  CarryVariety := Stones [Settings [BlockLength]].Variety;

  { And copy all varieties one down }
  if Settings [BlockLength] <> 1 then begin
    for StoneNr := Settings [BlockLength] downto 2 do begin
      Stones [StoneNr].Variety := Stones [StoneNr - 1].Variety;
    end; { for }
  end; { then }

  { And put in the carried variety }
  Stones [1].Variety := CarryVariety;
end; { TurnDown }

{ Gives the block a new coordinate }
procedure TBlock.MakeNewCoord;

var
  StoneNr : byte;

begin
  { Fill in the block coordinate }
  Position.X := Settings [FieldWidth] div 2 + 1;
  Position.Y := 0;

  { And all separate stone coordinates }
  for StoneNr := 1 to Settings [BlockLength] do begin
    with Stones [StoneNr].Coord do begin
      X := Position.X;
      Y := Position.Y + StoneNr;
    end; { with }
  end; { for }
end; { MakeNewCoord }

{ Makes new varieties in block }
procedure TBlock.MakeNewVarieties;

var
  StoneNr : byte;

begin
  { For every stone in block }
  for StoneNr := 1 to Settings [BlockLength] do begin
    Stones [StoneNr].Variety := Random (Settings [AmVarieties]) + 1;
    if Settings [Joker] <> 0 then begin
      if Random (Settings [Joker]) = 0 then begin
        { Make joker }
        Stones [StoneNr].Variety := JokerStoneNr;
      end; { then }
    end; { then }
  end; { for }
end; { MakeNewVarieties }

{ Makes a new block }
procedure TBlock.MakeNewBlock;

var
  StoneNr : byte;

begin
  MakeNewCoord;
  MakeNewVarieties;

  for StoneNr := 1 to Settings [BlockLength] do begin
    Stones [StoneNr].Sortof := NormalStone;
  end; { for }

  { And look if it is already stuck }
  Stuck := false;
  for StoneNr := 1 to Settings [BlockLength] do begin
    if Field.Occupied (Stones [StoneNr].Coord) then begin
      Stuck := true;
    end; { then }
  end; { for }
end; { MakeNewBlock }

{ Moves the block one down in the field }
procedure TBlock.MoveDown;

var
  NextCoord : TCoord;
  StoneNr   : byte;

begin
  { Make the coordinate under the block }
  NextCoord := Position;
  NextCoord.Y := NextCoord.Y + Settings [BlockLength] + 1;

  { Look if the coordinate under the block is occupied (if it is stuck) }
  Stuck := Field.Occupied (NextCoord);
  if Stuck then begin
    Exit;
  end; { then }

  { And if not so, move the block down }
  Position.Y := Position.Y + 1;
  for StoneNr := 1 to Settings [BlockLength] do begin
    Inc (Stones [StoneNr].Coord.Y);
  end; { for }
end; { MoveDown }

{ Moves the block one left }
procedure TBlock.MoveLeft;

var
  NextCoord : TCoord;
  StoneNr   : byte;

begin
  { Block is not stuck by definition }
  Stuck := false;

  { Make the first (upper) coordinate left of the block }
  NextCoord := Position;
  if NextCoord.X <> 0 then begin
    Dec (NextCoord.X);
  end { then }
  else begin
    Stuck := true;
    Exit;
  end; { else }

  { And look if the coordinates left of the block are occupied }
  for StoneNr := 1 to Settings [BlockLength] do begin
    NextCoord := Stones [StoneNr].Coord;
    Dec (NextCoord.X);
    if Field.Occupied (NextCoord) then begin
      Stuck := true;
      Exit;
    end; { then }
  end; { for }

  { And if not so, move the block left }
  Dec (Position.X);
  for StoneNr := 1 to Settings [BlockLength] do begin
    Dec (Stones [StoneNr].Coord.X);
  end; { for }
end; { MoveLeft }

{ Moves the block one Right }
procedure TBlock.MoveRight;

var
  NextCoord : TCoord;
  StoneNr   : byte;

begin
  { Block is not stuck by definition }
  Stuck := false;

  { Make the first (upper) coordinate right of the block }
  NextCoord := Position;
  if NextCoord.X <> Settings [FieldWidth] then begin
    Inc (NextCoord.X);
  end { then }
  else begin
    Stuck := true;
    Exit;
  end; { else }

  { And look if the coordinates left of the block are occupied }
  for StoneNr := 1 to Settings [BlockLength] do begin
    NextCoord := Stones [StoneNr].Coord;
    Inc (NextCoord.X);
    if Field.Occupied (NextCoord) then begin
      Stuck := true;
      Exit;
    end; { then }
  end; { for }

  { And if not so, move the block left }
  Inc (Position.X);
  for StoneNr := 1 to Settings [BlockLength] do begin
    Inc (Stones [StoneNr].Coord.X);
  end; { for }
end; { MoveRight }

{ Gets the stone at the specified StoneNr }
procedure TBlock.GetStone (StoneNr : byte; var Stone : TStone);

begin
  Stone := Stones [StoneNr];
end; { GetStone }






{ Field member functions -----------------------------------------------------}

{ Inserts the given Block in the field }
procedure TField.InsertBlock;

var
  BlockStone : TStone;
  StoneCoord : TCoord;
  StoneNr    : byte;

begin
  { For every stone in the block }
  for StoneNr := 1 to Settings [BlockLength] do begin
    { Get it's coordinate }
    BlockStone := Block.Stones [StoneNr];
    StoneCoord := BlockStone.Coord;

    { And copy the stone in the right fieldplace }
    AllStones [StoneCoord.X, StoneCoord.Y] := BlockStone;
  end; { for }
end; { InsertBlock }

{ Tells wether the given coordinate is occupied }
function TField.Occupied (Coord : TCoord) : boolean;

var
  FieldStone : TStone;

begin
  { Look if the given coordinate is out of the field boundaries }
  if (Coord.Y > Settings [FieldLength]) or (Coord.X > Settings [FieldWidth]) or
     (Coord.X < 1) or (Coord.Y < 1) then begin
    Occupied := true;
    Exit;
  end; { then }

  { Refer to the corresponding stone in the field }
  FieldStone := AllStones [Coord.X, Coord.Y];

  { And look if the stone is present }
  Occupied := (FieldStone.Sortof <> NoStone);
end; { Occupied }

{ Clears the field }
procedure TField.ClearField;

var
  XCoord,
  YCoord     : byte;
  FieldStone : TStone;

begin
  { For every stone in the field }
  for XCoord := 1 to Settings [FieldWidth] do begin
    for YCoord := 1 to Settings [FieldLength] do begin
      { Make the right empty stone }
      FieldStone.Coord.X := XCoord;
      FieldStone.Coord.Y := YCoord;
      FieldStone.Sortof := NoStone;

      { And copy the stone in the field }
      AllStones [XCoord, YCoord] := FieldStone;
    end; { for }
  end; { for }
end; { ClearField }

{ Gets the stone at the specified coordinates }
procedure TField.GetStone (Coord : TCoord; var Stone : TStone);

begin
  Stone := AllStones [Coord.X, Coord.Y];
end; { GetStone }



{ Returns how many stones were deleted in full rows }
function TField.CheckRows : word;

var
  NextCoord,
  FieldCoord    : TCoord;
  AmDeleted     : word;
  SearchVariety : byte;
  Stone         : TStone;
  StonesDeleted,
  RowFinished   : boolean;

  { Checks the full row to one side, from the StartCoord on }
  procedure CheckOneSide (DX, DY : integer; SearchVariety : byte;
                          StartCoord : TCoord);

  var
    NextCoord : TCoord;
    StoneNr,
    AmInRow   : byte;

    { Moves the NextCoord to the next coordinate }
    procedure GoNextCoord;

    begin
      case DX of
        +1 : Inc (NextCoord.X);
        -1 : Dec (NextCoord.X);
      end; { case }
      case DY of
        +1 : Inc (NextCoord.Y);
        -1 : Dec (NextCoord.Y);
      end; { case }
    end; { GoNextCoord }

    { Checks whether the nextcoord is still within the field boundaries }
    function NextCoordInField : boolean;

    begin
      NextCoordInField := (NextCoord.X <= Settings [FieldWidth]) and
                          (NextCoord.X > 0) and
                          (NextCoord.Y <= Settings [FieldLength]) and
                          (NextCoord.Y > 0);
    end; { NextCoordInField }

  begin
    { Set the coordinate references }
    NextCoord := FieldCoord;

    { And if the to-search stone is a joker, }
    while (SearchVariety = JokerStoneNr) and NextCoordInField do begin
      { Then take the next stone in the row instead }
      GoNextCoord;
      GetStone (NextCoord, Stone);
      SearchVariety := Stone.Variety;
    end; { while }

    { And start checking }
    NextCoord := FieldCoord;
    GoNextCoord;
    AmInRow := 1;
    RowFinished := false;
    while NextCoordInField and not RowFinished do begin
      GetStone (NextCoord, Stone);
      if (Stone.Sortof <> NoStone) and ((Stone.Variety = SearchVariety) or (Stone.Variety = JokerStoneNr)) then begin
        Inc (AmInRow);
      end { then }
      else begin
        RowFinished := true;
      end; { else }
      GoNextCoord;
    end; { while }

    { Look if there are enough of 'em in the row }
    if AmInRow >= Settings [RowLength] then begin
      { Yes -> delete them all }
      NextCoord := FieldCoord;
      for StoneNr := 1 to AmInRow do begin
        AllStones [NextCoord.X, NextCoord.Y].SortOf := DeletedStone;
        GoNextCoord;
      end; { for }
      StonesDeleted := true;
    end; { then }
  end; { CheckOneSide }

begin
  { No stones deleted so far }
  StonesDeleted := false;

  { Start checki2ng on every coordinate }
  for FieldCoord.X := 1 to Settings [FieldWidth] do begin
    for FieldCoord.Y := 1 to Settings [FieldLength] do begin
      { Get the stone at that position }
      Field.GetStone (FieldCoord, Stone);

      { And if there is a stone there }
      if Stone.Sortof <> NoStone then begin
        { Get the variety to search for }
        SearchVariety := Stone.Variety;

        { Check to the right }
        CheckOneSide (+1, 0, SearchVariety, FieldCoord);

        { To the right-down }
        CheckOneSide (+1, -1, SearchVariety, FieldCoord);

        { To the down }
        CheckOneSide (0, -1, SearchVariety, FieldCoord);

        { To the left-down }
        CheckOneSide (-1, -1, SearchVariety, FieldCoord);

        { To the left }
        CheckOneSide (-1, 0, SearchVariety, FieldCoord);

        { To the left-up }
        CheckOneSide (-1, +1, SearchVariety, FieldCoord);

        { To the up }
        CheckOneSide (0, +1, SearchVariety, FieldCoord);

        { To the right-up }
        CheckOneSide (+1, +1, SearchVariety, FieldCoord);
      end; { then }
    end; { for }
  end; { for }

  { And if stones are deleted }
  if StonesDeleted then begin
    { Count the amount of them and rearrange other stones }
    AmDeleted := 0;
    for FieldCoord.Y := 1 to Settings [FieldLength] do begin
      for FieldCoord.X := 1 to Settings [FieldWidth] do begin
        { Get the corresponding stone }
        GetStone (FieldCoord, Stone);

        { And look if it is a deleted one }
        if Stone.Sortof = DeletedStone then begin
          { Found an extra one }
          AmDeleted := AmDeleted + 1;

          { Move upper row one down }
          NextCoord := FieldCoord;
          while NextCoord.Y <> 1 do begin
            Dec (NextCoord.Y);
            GetStone (NextCoord, Stone);
            AllStones [NextCoord.X, NextCoord.Y + 1].Variety := Stone.Variety;
            AllStones [NextCoord.X, NextCoord.Y + 1].Sortof := Stone.Sortof;
          end; { while }

          { And delete the upper stone }
          AllStones [NextCoord.X, 1].Sortof := NoStone;
        end; { then }
      end; { for }
    end; { for }

    { And return the amount of deleted stones }
    CheckRows := AmDeleted;
  end { then }
  else begin
    { No stones deleted }
    CheckRows := 0;
  end; { else }
end; { CheckRows }

{ Makes full rows in the bottom of the field }
procedure TField.MakeFullRows;

var
  AmDeleted : word;
  RowNr     : byte;
  Coord     : TCoord;

begin
  { Look if full rows need to be made }
  if FullRows <> 0 then begin
    { Make a first try }
    for Coord.Y := Settings [FieldLength] downto Settings [FieldLength] - FullRows + 1 do begin
      for Coord.X := 1 to Settings [FieldWidth] do begin
        AllStones [Coord.X, Coord.Y].Variety := Random (Settings [AmVarieties]) + 1;
        AllStones [Coord.X, Coord.Y].SortOf := NormalStone;
        AllStones [Coord.X, Coord.Y].Coord := Coord;
      end; { for }
    end; { for }

    { And keep trying until all rows full at random }
    repeat
      { Check on full rows }
      AmDeleted := CheckRows;

      { And if stones are deleted, refill their places }
      if AmDeleted <> 0 then begin
        for Coord.Y := Settings [FieldLength] downto Settings [FieldLength] - FullRows + 1 do begin
          for Coord.X := 1 to Settings [FieldWidth] do begin
            if AllStones [Coord.X, Coord.Y].SortOf = NoStone then begin
              { Try another stone }
              AllStones [Coord.X, Coord.Y].Variety := Random (Settings [AmVarieties]) + 1;
              AllStones [Coord.X, Coord.Y].SortOf := NormalStone;
              AllStones [Coord.X, Coord.Y].Coord := Coord;
            end; { then }
          end; { for }
        end; { for }
      end; { then }
    until AmDeleted = 0;
  end; { then }
end; { MakeFullRows }


begin
  Randomize;
  Field.ClearField;
end. { CF_Play }