{-----------------------------------------------------------------------------

 CF_Scor : score and hiscore list interface of the game ChiFu.

 Consists of:

 Types:
 -TEntry     : single hiscore entry definition
 -TScoreList : single hiscore list definition
 -TScores    : score and hiscore list interface definition
   Variables:
   -Score    : the score of the player
   -HiScores : the hiscore list

   Routines:
   -Init          : initializes the scores
   -AddScore      : adds the score to the hiscore list, if nescesary
   -IsHiScore     : tells whether the score belongs in the hiscore list or not
   -ClearHiScores : clears the hiscore list
   -ShowScore     : shows the score on screen
   -ShowHiScores  : shows the hiscore list

 Constants:
 -NameLength : standard maximum name length in hiscore entries
 -MaxScores  : the maximum amount of scores in the hiscore list

 Variables:
 -Scores : scores and hiscore lists interface

 Routines:

 Defaults:
 -Initializes the score and hiscore interface

-----------------------------------------------------------------------------}
unit CF_Scor;

interface

uses
  CF_Set,
  CF_Grph,
  CF_KIO;

const
  NameLength = 15;
  MaxScores  = 15;

type
  { Single player entry in score list }
  TEntry = record
             Name  : string [NameLength];
             Score : longint;
           end; { record }

  { Entire hiscore list }
  TScoreList = array [0..MaxScores] of TEntry;

  { Score and hiscore interface }
  TScores = object
              public
                { The score of the player }
                Score : longint;

                { The hiscore list }
                HiScores : TScoreList;

                { Initializes the scores }
                constructor Init;

                { Adds the score to the hiscore list, if nescesary }
                procedure AddScore;

                { Tells whether the score belongs in the hiscore list or not }
                function IsHiscore : boolean;

                { Clears the hiscore list }
                procedure ClearHiScores;

                { Shows the score on screen }
                procedure ShowScore;

                { Shows the hiscore list }
                procedure ShowHiscores;
              private

            end; { object }

var
  Scores : array [TPlayMode] of TScores;

implementation

const
  { A standard empty hiscore list entry }
  EmptyEntry : TEntry = (Name:''; Score:-1);

{ Initializes the scores }
constructor TScores.Init;

begin
  Score := 0;
  ClearHiScores;
end; { Init }

{ Tells whether the score belongs in the hiscore list or not }
function TScores.IsHiscore : boolean;

begin
  { Look if the score can enter the list }
  if Score < HiScores [MaxScores].Score then begin
    { Score cannot enter }
    IsHiscore := false;
  end { then }
  else begin
    IsHiscore := true;
  end; { else }
end; { IsHiscore }

{ Adds the score to the hiscore list, if nescesary }
procedure TScores.AddScore;

var
  EntryPlace,
  ScorePlace : integer;
  NumberText : string;

begin
  { Look where the score would end up in the list }
  ScorePlace := MaxScores;
  while (ScorePlace >= 0) and (Score >= HiScores [ScorePlace].Score) do begin
    Dec (ScorePlace);
  end; { while }
  EntryPlace := ScorePlace + 1;

  { Rearrange the list }
  for ScorePlace := MaxScores downto EntryPlace + 1 do begin
    HiScores [ScorePlace] := HiScores [ScorePlace - 1];
  end; { for }

  { Show the current hiscore list }
  ShowHiScores;

  { Show to enter name }
  Screen.ScrollDown;
  Screen.ScrollDown;
  Str (Score, NumberText);
  Screen.ShowText ('Your final score is ' + NumberText + '.');
  Screen.ScrollDown;
  Screen.ScrollDown;
  Screen.ShowText ('This means that you made it into the hiscore list');
  Screen.ScrollDown;
  Str (EntryPlace, NumberText);
  Screen.ShowText ('at number ' + NumberText + '!');
  Screen.ScrollDown;
  Screen.ScrollDown;
  Screen.ShowText ('Please enter your name; 15 letters max.');

  { And go to the corresponding place in the hiscore list }
  Screen.CursorPos := Screen.Home;
  Screen.ScrollDown;
  Screen.ScrollDown;
  Screen.ScrollDown;
  for ScorePlace := 1 to EntryPlace do begin
    Screen.ScrollDown;
  end; { for }
  Screen.ScrollRight;
  Screen.ScrollRight;
  Screen.ScrollRight;
  Screen.ScrollRight;

  { Empty the line already present }
  Screen.ShowText ('                       ');

  { Let the user enter his name }
  Screen.ShowCursor;
  KeyBoard.CaseOn := true;
  HiScores [EntryPlace].Name  := Keyboard.GetText;
  KeyBoard.CaseOn := false;

  { And put in the new score }
  HiScores [EntryPlace].Score := Score;
  Screen.HideCursor;
end; { AddScore }

{ Clears the hiscore list }
procedure TScores.ClearHiScores;

var
  EntryNr : byte;

begin
  for EntryNr := 0 to MaxScores do begin
    HiScores [EntryNr] := EmptyEntry;
  end; { for }
end; { ClearHiScores }

{ Shows the score on screen }
procedure TScores.ShowScore;

var
  TextLine : string;

begin
  Screen.CursorPos := Screen.ScorePos;
  Str (Score, TextLine);
  Screen.ShowText ('Score: ' + TextLine + ' ');
end; { ShowScore }

{ Shows the hiscore list }
procedure TScores.ShowHiScores;

var
  EntryNr  : byte;
  StrText,
  TextLine : string;

begin
  Screen.ClearScreen;
  Screen.CursorPos := Screen.Home;
  Screen.ShowText ('Chi Fu AllTime Hiscore List for Play Mode ' +
                   PlayModeNames [PlayMode] + ':');

  Screen.ScrollDown;
  Screen.ScrollDown;
  Screen.ScrollDown;

  for EntryNr := 0 to MaxScores do begin
    Str (EntryNr, TextLine);
    if Length (TextLine) = 1 then begin
      TextLine := ' ' + TextLine;
    end; { then }
    if HiScores [EntryNr].Score = EmptyEntry.Score then begin
      TextLine := TextLine + '; ---';
    end { then }
    else begin
      TextLine := TextLine + '; ' + HiScores [EntryNr].Name;
      Str (HiScores [EntryNr].Score, StrText);
      TextLine := TextLine + '; ' + StrText;
    end; { else }
    Screen.ShowText (TextLine);
    Screen.ScrollDown;
  end; { for }

  Screen.ScrollDown;
  Screen.ScrollDown;
  Screen.ScrollDown;
end; { ShowHiScores }

begin
  Scores [Large].Init;
  Scores [Normal].Init;
  Scores [Small].Init;
end. { CF_Scor }