{-----------------------------------------------------------------------------

 MakeSound : enables the users to make their own sounds with the game Chi Fu

-----------------------------------------------------------------------------}
program MakeSnd;

uses
  Graph,
  WinDos,
  Strings,
  Crt;

var
  Changed,
  Finished     : boolean;
  UserChoice   : char;

const
  { Maximum amount of elements in the tune }
  MaxTuneSize = 200;

type
  { A single tune element }
  TTuneElem = record
                Freq  : word;  { Frequency of the note }
                Delay : byte;  { Delay for the note in ms }
              end; { record }

  { And a complete tune }
  TTune = record
            Length   : byte;       { Total length of tune in elements }
            Duration : word;       { Total duration of tune in ms }
            Elements : array [1..MaxTuneSize] of TTuneElem;
                                   { All tune elements }
          end; { record }

  { A tune file on disk }
  TTuneFile = file of TTuneElem;

var
  { The tune itself }
  Tune : TTune;






{ Sound part ----------------------------------------------------------------}

{ Plays the tune }
procedure PlayTune;

var
  TuneNr,
  TuneElemNr : byte;

begin
  { Play the tune until the end }
  for TuneElemNr := 1 to Tune.Length do begin
    { Set the speaker to the right amount of Hz }
    Sound (Tune.Elements [TuneElemNr].Freq);

    { Wait for the specified amount of time }
    Delay (Tune.Elements [TuneElemNr].Delay);
  end; { for }

  { And kill the last note again }
  NoSound;
end; { PlayTune }







{ Graphical part ------------------------------------------------------------}

const
  YSize = 350;

var
  MaxFreq,
  MinFreq : word;
  XScale,
  YScale  : real;

{ Initializes the screen into graphics-mode }
procedure InitializeScreen;

var
  Driver,
  Mode : integer;

begin
  { Put screen in graphics-mode }
  Driver := Detect;
  InitGraph (Driver, Mode, '.');
end; { Initialize }

{ Sets the scales to the tune currently in use }
procedure SetScales;

var
  ElemNr : byte;

begin
  { Scan for the highest/lowest frequencies }
  MaxFreq := 0;
  MinFreq := $FFFF;
  for ElemNr := 1 to Tune.Length do begin
    if MinFreq > Tune.Elements [ElemNr].Freq then begin
      MinFreq := Tune.Elements [ElemNr].Freq;
    end; { then }
    if MaxFreq < Tune.Elements [ElemNr].Freq then begin
      MaxFreq := Tune.Elements [ElemNr].Freq;
    end; { then }
  end; { for }
  XScale := GetMaxX / Tune.Duration;
  YScale := 1;
  if MaxFreq <> MinFreq then begin
    YScale := YSize / (MaxFreq - MinFreq);
  end; { then }
end; { SetScales }

{ Gives the screen coordinates of the given tune offset }
procedure MakeScreenCoord (Freq, Duration : word; var X, Y : word);

begin
  X := Trunc (Duration * XScale);
  Y := YSize - Trunc (YScale * (Freq - MinFreq)) + 20;
end; { MakeScreenCoord }

{ Shows the main screen }
procedure ShowMainScreen;

begin
  SetColor (15);
  OutTextXY (0, 390, 'S          : save tune');
  OutTextXY (0, 400, 'P          : play tune');
  OutTextXY (0, 410, 'Esc        : end program');
  OutTextXY (0, 420, 'PgUp/Down  : move left/right');
  OutTextXY (0, 430, 'Ins/Del    : insert/delete note');
  OutTextXY (0, 440, 'Left/Right : in-/decrease the duration');
  OutTextXY (0, 450, 'Up/Down    : in-/decrease the frequency');
end; { ShowMainScreen }

{ Shows the tune on screen }
procedure ShowTune;

var
  Color      : byte;
  X,
  Y,
  ElemNr,
  Duration   : word;
  NumberText : string;

begin
  { Show the min/max freq. measures }
  SetColor (15);
  ClearViewPort;
  Str (MaxFreq, NumberText);
  OutTextXY (0, 20, NumberText + ' Hz');
  Str (MinFreq, NumberText);
  OutTextXY (0, 20 + YSize, NumberText + ' Hz');

  { Start with light green }
  Color := 10;
  SetColor (Color);

  { Calculate the scales }
  SetScales;

  { And the rest of the tune }
  Duration := 0;
  for ElemNr := 1 to Tune.Length do begin
    { Show the line }
    MakeScreenCoord (Tune.Elements [ElemNr].Freq, Duration, X, Y);
    MoveTo (X, Y);
    Duration := Duration + Tune.Elements [ElemNr].Delay;
    MakeScreenCoord (Tune.Elements [ElemNr].Freq, Duration, X, Y);
    LineTo (X, Y);

    { And change the color to the other one }
    if Color = 10 then begin
      Color := 14;
    end { then }
    else begin
      Color := 10;
    end; { else }
    SetColor (Color);
  end; { for }

  { And show the total duration }
  Str (Tune.Duration, NumberText);
  SetColor (15);
  OutTextXY (0, 0, 'Duration of tune: ' + NumberText + ' ms');
  ShowMainScreen;
end; { ShowTune }

{ Shows the cursor and deletes the previous one }
procedure ShowCursor (X, Y, PrevX, PrevY : word);

begin
  { Delete previous one }
  SetColor (0);
  Rectangle (PrevX - 3, PrevY - 3, PrevX + 3, PrevY + 3);

  { And show the new one }
  SetColor (15);
  Rectangle (X - 3, Y - 3, X + 3, Y + 3);
end; { ShowCursor }






{ File part -----------------------------------------------------------------}

var
  TuneFile : TTuneFile;
  FileName : string;

{ Sets the file name to the one from the command line }
procedure SetFileName;

var
  CmdLine : PChar;

begin
  { Get the file name to use }
  GetMem (CmdLine, 255);
  GetArgStr (CmdLine, 1, 250);
  if CmdLine [0] = #00 then begin
    { No name given -> tell so }
    WriteLn ('No file name specified.');
    WriteLn ('Have to give name of file to process.');
    Halt (1);
  end; { then }
  FileName := StrPas (CmdLine);
  Assign (TuneFile, FileName);
end; { SetFileName }

{ Saves the tune to disk }
procedure SaveTune;

var
  ElemNr : byte;

begin
  { And start saving it to disk }
  Rewrite (TuneFile);
  if Tune.Length <> 0 then begin
    for ElemNr := 1 to Tune.Length do begin
      Write (TuneFile, Tune.Elements [ElemNr]);
    end; { for }
  end; { then }
  Close (TuneFile);
end; { SaveTune }

{ Makes a new (standard) data file }
procedure MakeNewFile;

begin
  Tune.Elements [1].Freq := 2500;
  Tune.Elements [1].Delay := 5;
  Tune.Length := 1;
  SaveTune;
end; { MakeNewFile }

{$I-}

{ FileExists : looks whether the given file exists or not }
function FileExists : boolean;

var
  TryFile : text;

begin
  Assign (TryFile, FileName);
  Reset (TryFile);
  FileExists := (IOResult = 0);
end; { FileExists }

{$I+}

{ Gets the tune from file }
procedure ReadTune;

var
  TuneElemNr : byte;
  ReadAway   : TTuneElem;

begin
  { Get the length of the tune }
  Reset (TuneFile);
  Tune.Length := 0;
  Tune.Duration := 0;
  while not Eof (TuneFile) do begin
    Inc (Tune.Length);
    Read (TuneFile, Tune.Elements [Tune.Length]);
    Tune.Duration := Tune.Duration + Tune.Elements [Tune.Length].Delay;
  end; { while }
  Close (TuneFile);
end; { ReadTune }








{ Keyboard part -------------------------------------------------------------}

const
  Esc    : char = #27;
  BckSpc : char = #08;
  Del    : char = #20;
  Ins    : char = #01;

  Up    : char = #14;
  Down  : char = #15;
  Left  : char = #16;
  Right : char = #17;

  PgUp   : char = #02;
  PgDown : char = #03;

{ GetKey : lets the user enter a key on the keyboard }
function GetKey : char;

var
  InChar : char;

  { LowerChar : turns the given char to lower-case }
  function LowerChar (InChar : char) : char;

  begin
    if (InChar >= 'A') and (InChar <= 'Z') then begin
      LowerChar := char (byte (InChar) + 32);
    end { then }
    else begin
      LowerChar := InChar;
    end; { else }
  end; { LowerChar }

begin
  repeat until KeyPressed;
  InChar := ReadKey;
  if InChar = #00 then begin
    InChar := ReadKey;
    case InChar of
      #72 : InChar := Up;
      #75 : InChar := Left;
      #77 : InChar := Right;
      #80 : InChar := Down;

      #82 : InChar := Ins;
      #83 : InChar := Del;
      #73 : InChar := PgUp;
      #81 : InChar := PgDown;
    end; { case }
  end; { then }

  GetKey := LowerChar (InChar);
end; { GetKey }

{ GetSelectedKey : enables the user to enter one of the selected keys and
                   gives back this key }
function GetSelectedKey (Selection : string) : char;

var
  InChar     : char;
  OptionNr   : byte;
  RightInput : boolean;

begin
  RightInput := false;
  repeat
    InChar := GetKey;
    for OptionNr := 1 to Length (Selection) do begin
      if InChar = Selection [OptionNr] then begin
        RightInput := true;
      end; { then }
    end; { for }
  until RightInput;

  GetSelectedKey := InChar;
end; { GetSelectedKey }








{ Cursor and user modification part -----------------------------------------}

var
  PrevX,
  PrevY,
  X,
  Y,
  Duration,
  ElemNr : word;

{ Puts a new pixel of the specified color in the picture }
procedure PutNewPixel;

begin
  ShowCursor (X, Y, PrevX, PrevY);
end; { PutNewPixel }

{ Updates the cursor place }
procedure UpdateCursor;

begin
  PrevX := X;
  PrevY := Y;
end; { UpdateCursor }

{ Increases the frequency }
procedure IncFreq;

begin
  with Tune.Elements [ElemNr] do begin
    if Freq < 20000 then begin
      Inc (Freq, 10);
      SetScales;
      ShowTune;
      MakeScreenCoord (Freq, Duration, X, Y);
      ShowCursor (X, Y, PrevX, PrevY);
      UpdateCursor;
    end; { then }
  end; { with }
end; { IncFreq }

{ Decreases the frequncy }
procedure DecFreq;

begin
  with Tune.Elements [ElemNr] do begin
    if Freq > 10 then begin
      Dec (Freq, 10);
      SetScales;
      ShowTune;
      MakeScreenCoord (Freq, Duration, X, Y);
      ShowCursor (X, Y, PrevX, PrevY);
      UpdateCursor;
    end; { then }
  end; { with }
end; { DecFreq }

{ Increases the duration }
procedure IncDuration;

begin
  with Tune.Elements [ElemNr] do begin
    if Delay < 25 then begin
      Inc (Delay);
      Inc (Tune.Duration);
      SetScales;
      ShowTune;
      MakeScreenCoord (Freq, Duration, X, Y);
      ShowCursor (X, Y, PrevX, PrevY);
      UpdateCursor;
    end; { then }
  end; { with }
end; { IncDuration }

{ Decreases the duration }
procedure DecDuration;

begin
  with Tune.Elements [ElemNr] do begin
    if Delay > 1 then begin
      Dec (Delay);
      Dec (Tune.Duration);
      SetScales;
      ShowTune;
      MakeScreenCoord (Freq, Duration, X, Y);
      ShowCursor (X, Y, PrevX, PrevY);
      UpdateCursor;
    end; { then }
  end; { with }
end; { DecDuration }

{ Moves the cursor left }
procedure MoveLeft;

begin
  if ElemNr > 1 then begin
    Dec (ElemNr);
    Duration := Duration - Tune.Elements [ElemNr].Delay;
    MakeScreenCoord (Tune.Elements [ElemNr].Freq, Duration, X, Y);
    ShowCursor (X, Y, PrevX, PrevY);
    UpdateCursor;
  end; { then }
end; { MoveLeft }

{ Moves the cursor right }
procedure MoveRight;

begin
  if ElemNr < Tune.Length then begin
    Duration := Duration + Tune.Elements [ElemNr].Delay;
    Inc (ElemNr);
    MakeScreenCoord (Tune.Elements [ElemNr].Freq, Duration, X, Y);
    ShowCursor (X, Y, PrevX, PrevY);
    UpdateCursor;
  end; { then }
end; { MoveRight }

{ Adds a new note }
procedure AddNote;

var
  NewElemNr : byte;

begin
  { Look if adding is allowed }
  if Tune.Length = MaxTuneSize then begin
    { No -> forget about it }
    Exit;
  end; { then }

  { Move the next notes up and the new one will be a copy of the old one }
  for NewElemNr := Tune.Length downto ElemNr do begin
    Tune.Elements [NewElemNr + 1] := Tune.Elements [NewElemNr];
  end; { for }

  { And adjust the length of the tune }
  Tune.Duration := Tune.Duration + Tune.Elements [ElemNr].Delay;
  Inc (Tune.Length);

  { And show the new situation }
  SetScales;
  ShowTune;
  MakeScreenCoord (Tune.Elements [ElemNr].Freq, Duration, X, Y);
  ShowCursor (X, Y, PrevX, PrevY);
  UpdateCursor;
end; { AddNote }

{ Deletes a note }
procedure DelNote;

var
  NewElemNr : byte;

begin
  { Look if adding is allowed }
  if Tune.Length = 1 then begin
    { No -> forget about it }
    Exit;
  end; { then }

  { Adjust the length of the tune }
  Tune.Duration := Tune.Duration - Tune.Elements [ElemNr].Delay;

  { Move the next notes up one place }
  for NewElemNr := ElemNr to Tune.Length do begin
    Tune.Elements [NewElemNr] := Tune.Elements [NewElemNr + 1];
  end; { for }

  { And show the new situation }
  Dec (Tune.Length);
  SetScales;
  ShowTune;
  if ElemNr > Tune.Length then begin
    Dec (ElemNr);
    Duration := Duration - Tune.Elements [ElemNr].Delay;
  end; { then }
  MakeScreenCoord (Tune.Elements [ElemNr].Freq, Duration, X, Y);
  ShowCursor (X, Y, PrevX, PrevY);
  UpdateCursor;
end; { DelNote }




{ Main program module -------------------------------------------------------}

begin
  { Set the file name and file to use }
  SetFileName;

  { Look if the file exists }
  if not FileExists then begin
    WriteLn ('File ''', FileName, ''' doesn''t exist.');
    WriteLn ('Make this file? (Y/N)');
    UserChoice := GetSelectedKey ('yn');
    if UserChoice = 'n' then begin
      Halt (2);
    end; { then }
    InitializeScreen;
    MakeNewFile;
  end; { then }

  { And read the tune }
  ReadTune;

  { Initialize the screen }
  InitializeScreen;
  PrevX := 600;
  PrevY := 0;
  ElemNr := 1;
  Duration := 0;
  SetScales;
  MakeScreenCoord (Tune.Elements [ElemNr].Freq, Duration, X, Y);

  { Show the main screen }
  ShowMainScreen;

  { And show the rest }
  ShowTune;
  ShowCursor (X, Y, PrevX, PrevY);

  Finished := false;
  Changed := false;
  repeat
    { Wait for user approval }
    UserChoice := GetSelectedKey ('sp' + Esc + Up + Down + Left + Right +
                                  Ins + Del + BckSpc + PgUp + PgDown);

    if UserChoice = Esc then begin
      Finished := true;
    end { then }
    else if UserChoice = 's' then begin
      SaveTune;
      Changed := false;
    end { else then }
    else if UserChoice = 'p' then begin
      PlayTune;
    end { else then }
    else if UserChoice = Up then begin
      IncFreq;
      Changed := true;
    end { else then }
    else if UserChoice = Down then begin
      DecFreq;
      Changed := true;
    end { else then }
    else if UserChoice = Right then begin
      IncDuration;
      Changed := true;
    end { else then }
    else if UserChoice = Left then begin
      DecDuration;
      Changed := true;
    end { else then }
    else if UserChoice = PgUp then begin
      MoveLeft;
    end { else then }
    else if UserChoice = PgDown then begin
      MoveRight;
    end { else then }
    else if UserChoice = Ins then begin
      AddNote;
      Changed := true;
    end { else then }
    else if (UserChoice = Del) or (UserChoice = BckSpc) then begin
      DelNote;
      Changed := true;
    end; { else then }
  until Finished;

  { Look if user changed picture and didn't save }
  if Changed then begin
    SetColor (15);
    OutTextXY (0, 150, 'Tune has been changed. Save to disk before exit? (Y/N)');
    UserChoice := GetSelectedKey ('yn');
    if UserChoice = 'y' then begin
      { Yes -> do so }
      SaveTune;
    end; { then }
  end; {then }

  RestoreCrtMode;
end. { MakeSnd }